program kmeans
  implicit none

  ! Number of cluster
  !integer, parameter :: CLUSTER_SIZE = 5
  integer, parameter :: CLUSTER_SIZE = 53
  ! Number of points
  !integer, parameter :: POINTS_SIZE = 100000
  integer, parameter :: POINTS_SIZE = 1000000
  ! Dimension of position
  integer, parameter :: POINTS_DIMENSION = 2
  ! Max iteration for convergence
  !integer, parameter :: MAX_ITER = 100
  ! Size of image output
  integer, parameter :: X_MAX = 640
  integer, parameter :: Y_MAX = 480
  ! Kind pour double precision
  integer, parameter :: dp = selected_real_kind(15,307)

  type :: point_array
    ! position of the points
    real(kind=dp), dimension(:,:),allocatable :: position
    ! centroid  id  associated
    integer, dimension(:), allocatable :: centroid_id
  end type

  type :: centroid_array
    ! position of center of centroid
    real(kind=dp), dimension(:,:), allocatable :: position
    ! accumulate position (for computation of new position)
    real(kind=dp), dimension(:,:), allocatable :: acc_position
    ! accumulate number of points
    integer, dimension(:), allocatable :: acc_points
    ! cte for dist
    real(kind=dp), dimension(:), allocatable :: halfdot
    ! color for image output
    integer,dimension(:,:), allocatable :: color
  end type

  type(point_array) :: tabpoints
  type(centroid_array) :: tabcentroids
  integer :: iter,iterk,ufile,iterc,row,col, nbiter
  real :: alea
  real :: start,finish
  logical :: again
  real(kind=dp) :: distmin, dist
  integer :: idmin
  integer, dimension(:,:,:), allocatable :: img

  ! Random init points
  allocate(tabpoints%position(POINTS_SIZE,POINTS_DIMENSION))
  allocate(tabpoints%centroid_id(POINTS_SIZE))
  do iter=1,POINTS_SIZE
    call random_number(alea)
    tabpoints%position(iter,1) = alea
    call random_number(alea)
    tabpoints%position(iter,2) = alea
  end do
  tabpoints%centroid_id = -1

  ! Init centroid (choose randomly points for centroids)
  allocate(tabcentroids%position(CLUSTER_SIZE,POINTS_DIMENSION))
  allocate(tabcentroids%acc_position(CLUSTER_SIZE,POINTS_DIMENSION))
  allocate(tabcentroids%acc_points(CLUSTER_SIZE))
  allocate(tabcentroids%color(CLUSTER_SIZE,3))
  allocate(tabcentroids%halfdot(CLUSTER_SIZE))
  do iter=1, CLUSTER_SIZE
    ! Choose randomly a point
    call random_number(alea)
    tabcentroids%position(iter,:) = tabpoints%position(1+int(alea*(POINTS_SIZE-1)),:)
    tabcentroids%halfdot(iter) = 0.5_dp*(tabcentroids%position(iter,1)*&
     tabcentroids%position(iter,1)+tabcentroids%position(iter,2)*&
     tabcentroids%position(iter,2))
    ! Define randomly a centroid color
    do iterc=1,3
      call random_number(alea)
      tabcentroids%color(iter,iterc) = (int(alea*256)+125)/2
    end do
  end do

  call cpu_time(start)
  nbiter = 0
  again = .true.
  do while(again)
    nbiter=nbiter+1
    again = .false.

    ! Init centroid
    tabcentroids%acc_position = 0._dp
    tabcentroids%acc_points = 0
    
    ! Find nearest centroid for all points
    do iter=1,POINTS_SIZE
      ! Starting with the last centroid
      distmin = tabcentroids%halfdot(CLUSTER_SIZE)-(tabcentroids%position(CLUSTER_SIZE,1)*&
       tabpoints%position(iter,1)+tabcentroids%position(CLUSTER_SIZE,2)*&
       tabpoints%position(iter,2))
      idmin = CLUSTER_SIZE
      ! Main loop
      do iterk=1, CLUSTER_SIZE-1
        dist = tabcentroids%halfdot(iterk)-(tabcentroids%position(iterk,1)*&
         tabpoints%position(iter,1)+tabcentroids%position(iterk,2)*&
         tabpoints%position(iter,2))
        if (dist < distmin) then
          distmin = dist
          idmin = iterk
        end if
      end do

      ! This point has a new nearest centroid ?
      if (idmin /= tabpoints%centroid_id(iter) ) then
        ! Yes so we do the while loop again
        tabpoints%centroid_id(iter) = idmin
        again = .true.
      end if

      ! Add point for gravity center computation of centroids
      tabcentroids%acc_position(idmin,1) = tabcentroids%acc_position(idmin,1)+ &
        tabpoints%position(iter,1)
      tabcentroids%acc_position(idmin,2) = tabcentroids%acc_position(idmin,2)+ &
        tabpoints%position(iter,2)
      tabcentroids%acc_points(idmin) = tabcentroids%acc_points(idmin)+1
    end do

    ! Position of centroids = gravity center of nearest points
    do iterk=1,CLUSTER_SIZE
      if (tabcentroids%acc_points(iterk) /= 0) then
        tabcentroids%position(iterk,1) = tabcentroids%acc_position(iterk,1)/ &
         tabcentroids%acc_points(iterk)
        tabcentroids%position(iterk,2) = tabcentroids%acc_position(iterk,2)/ &
         tabcentroids%acc_points(iterk)
        tabcentroids%halfdot(iterk) = 0.5_dp*(tabcentroids%position(iterk,1)*&
         tabcentroids%position(iterk,1)+tabcentroids%position(iterk,2)*&
         tabcentroids%position(iterk,2))
      end if
    end do
  end do
  call cpu_time(finish)
  print *, "Time ", finish-start,"s", nbiter,"its", nbiter/(finish-start),"it/s"

  ! Print position of centroids
  do iterk=1, CLUSTER_SIZE
    print *, iterk, tabcentroids%position(iterk,:)
  end do

  ! Write image
  allocate(img(3,X_MAX,Y_MAX))
  img = 0
  do iter=1,POINTS_SIZE
    col = int((X_MAX-1)*tabpoints%position(iter,1))+1
    row = int((Y_MAX-1)*tabpoints%position(iter,2))+1
    do iterc=1,3
      img(iterc,col,row) = tabcentroids%color(tabpoints%centroid_id(iter),iterc)
    end do
  end do
  do iterk=1,CLUSTER_SIZE
    col = int((X_MAX-1)*tabcentroids%position(iterk,1))+1
    row = int((Y_MAX-1)*tabcentroids%position(iterk,2))+1
    if ((col > 0) .and. (col < X_MAX) .and. (row > 0) .and. (row < Y_MAX) ) then
      do iterc=1,3
        img(iterc,col,row) = 255
      end do
    end if
  end do
  open(newunit=ufile,file="out.ppm")
  write(ufile,'(A2)') 'P6'
  write(ufile,'(i0," ",i0)') X_MAX,Y_MAX
  write(ufile,'(i0)') 255
  do row=1,Y_MAX
    do col=1,X_MAX
      write(ufile, '(3A1)', advance='no') img(1,col,row),img(2,col,row), &
       img(3,col,row)
    end do
  end do
  close(ufile)

end program
